@@ -1,28 +1,17 @@
 # Verilog - Verilog Perl Interface
-# $Id: Cell.pm 39061 2007-05-21 14:49:55Z wsnyder $
-# Author: Wilson Snyder <wsnyder@wsnyder.org>
-######################################################################
-#
-# Copyright 2000-2007 by Wilson Snyder.  This program is free software;
-# you can redistribute it and/or modify it under the terms of either the GNU
-# General Public License or the Perl Artistic License.
-# 
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-# 
+# See copyright, etc in below POD section.
 ######################################################################
 
 package Verilog::Netlist::Cell;
-use Class::Struct;
 
 use Verilog::Netlist;
 use Verilog::Netlist::Subclass;
+use vars qw($VERSION @ISA);
+use strict;
 @ISA = qw(Verilog::Netlist::Cell::Struct
 	Verilog::Netlist::Subclass);
-$VERSION = '2.380';
-use strict;
+
+$VERSION = '2.468';
 
 structs('new',
 	'Verilog::Netlist::Cell::Struct'
@@ -36,10 +25,12 @@ structs('new',
 	   submodname	=> '$', #'	# Which module it instantiates
 	   module	=> '$', #'	# Module reference
 	   params	=> '$', #'	# Textual description of parameters
+	   range	=> '$', #'	# Range of ranged instance
 	   _pins	=> '%',		# List of Verilog::Netlist::Pins
 	   byorder 	=> '$',		# True if Cell call uses order based pins
 	   # after link():
 	   submod	=> '$', #'	# Sub Module reference
+	   gateprim	=> '$', #'	# Primitive (and/buf/cmos etc), but not UDPs
 	   # system perl
 	   _autoinst	=> '$', #'	# Marked with AUTOINST tag
 	   ]);
@@ -57,6 +48,10 @@ sub delete {
 ######################################################################
 #### Methods
 
+sub logger {
+    my $self = shift;
+    return $self->netlist->logger;
+}
 sub netlist {
     my $self = shift;
     return $self->module->netlist;
@@ -64,53 +59,70 @@ sub netlist {
 
 sub _link_guts {
     my $self = shift;
-    if ($self->submodname()) {
-	my $name = $self->submodname();
-	my $sm = $self->netlist->find_module ($self->submodname());
-	if (!$sm) {
-	    $sm = $self->netlist->find_module ($self->netlist->remove_defines($self->submodname()));
+    # This function is HOT, keep simple
+    if (!$self->submod) {
+	if (my $name = $self->submodname) {
+	    my $netlist = $self->netlist;
+	    my $sm = $netlist->find_module_or_interface_for_cell($name);
+	    if (!$sm) {
+		my $name2 = $netlist->remove_defines($name);
+		$sm = $netlist->find_module_or_interface_for_cell($name2)
+		    if $name ne $name2;
+	    }
+	    if ($sm) {
+		$self->submod($sm);
+		$sm->is_top(0);
+	    }
 	}
-	$self->submod($sm);
-	$sm->is_top(0) if $sm;
-    }
-    foreach my $pinref ($self->pins) {
-	$pinref->_link();
     }
 }
 sub _link {
     my $self = shift;
+    # This function is HOT, keep simple
     $self->_link_guts();
+    if (!$self->submod && Verilog::Language::is_gateprim($self->submodname)) {
+	$self->gateprim(1);
+    }
     if (!$self->submod()
-	&& !$self->netlist->{_relink}
+	&& !$self->gateprim
 	&& !$self->module->is_libcell()
-	&& $self->netlist->{link_read}) {
+	&& $self->netlist->{link_read}
+	&& !$self->netlist->{_missing_submod}{$self->submodname}
+	) {
 	print "  Link_Read ",$self->submodname,"\n" if $Verilog::Netlist::Debug;
 	# Try 1: Direct filename
 	$self->netlist->read_file(filename=>$self->submodname, error_self=>0);
 	$self->_link_guts();
+	#
 	# Try 2: Libraries
 	if (!$self->submod()) {
 	    $self->netlist->read_libraries();
+	    $self->_link_guts();
 	}
-	$self->_link_guts();
 	# Try 3: Bitch about missing file
 	if (!$self->submod()) {
 	    $self->netlist->read_file(filename=>$self->submodname,
 				      error_self=>($self->netlist->{link_read_nonfatal} ? 0:$self));
 	}
-	# Got it; ask for another link
-	if ($self->submod()) {
-	    $self->netlist->{_relink} = 1;
+	# Still missing
+	if (!$self->submod()) {
+	    # Don't link this file again - speeds up if many common gate-ish missing primitives
+	    $self->netlist->{_missing_submod}{$self->submodname} = 1;
 	}
+	# Note if got it the new_module will add it to the _need_link list
+    }
+    # Link pins after module resolved, so don't do it multiple times if not found
+    foreach my $pinref ($self->pins) {
+	$pinref->_link();
     }
 }
 
 sub lint {
     my $self = shift;
-    if (!$self->submod() && !$self->netlist->{link_read_nonfatal}) {
-        $self->error ($self,"Module reference not found: ",$self->submodname(),,"\n");
+    if (!$self->submod() && !$self->gateprim && !$self->netlist->{link_read_nonfatal}) {
+        $self->error($self,"Module/Program/Interface reference not found: ",$self->submodname(),,"\n");
     }
-    if (!$self->netlist->{skip_pin_interconnect}) {
+    if ($self->netlist->{use_vars}) {
 	foreach my $pinref ($self->pins) {
 	    $pinref->lint();
 	}
@@ -119,7 +131,15 @@ sub lint {
 
 sub verilog_text {
     my $self = shift;
-    my @out = $self->submodname." ".$self->name." (";
+    my @out = $self->submodname;
+    if ($self->params) {
+	push @out, " #(".$self->params.")";
+    }
+    push @out, " ".$self->name;
+    if ($self->range) {
+	push @out, " ".$self->range;
+    }
+    push @out, " (";
     my $comma="";
     foreach my $pinref ($self->pins_sorted) {
 	push @out, $comma if $comma; $comma=", ";
@@ -153,16 +173,16 @@ sub new_pin {
     my $self = shift;
     # @_ params
     # Create a new pin under this cell
-    my $pinref = new Verilog::Netlist::Pin (cell=>$self, @_);
-    $self->portname($self->name) if !$self->name;	# Back Version 1.000 compatibility
-    $self->_pins ($pinref->name(), $pinref);
+    push @_, (cell=>$self);
+    my $pinref = new Verilog::Netlist::Pin(@_);
+    $self->_pins($pinref->name(), $pinref);
     return $pinref;
 }
 
 sub find_pin {
     my $self = shift;
     my $name = shift;
-    return $self->_pins($name);
+    return $self->_pins($name) || $self->_pins("\\".$name." ");
 }
 
 sub pins {
@@ -189,7 +209,7 @@ Verilog::Netlist::Cell - Instantiated cell within a Verilog Netlist
   use Verilog::Netlist;
 
   ...
-  my $cell = $module->find_cell ('cellname');
+  my $cell = $module->find_cell('cellname');
   print $cell->name;
 
 =head1 DESCRIPTION
@@ -212,6 +232,11 @@ passed to Verilog::Netlist::new for comments to be retained.
 
 Delete the cell from the module it's under.
 
+=item $self->gateprim
+
+True if the cell is a gate primitive instantiation (buf/cmos/etc), but not
+a UDP.
+
 =item $self->module
 
 Pointer to the module the cell is in.
@@ -232,6 +257,11 @@ List of Verilog::Netlist::Pin connections for the cell.
 
 List of name sorted Verilog::Netlist::Pin connections for the cell.
 
+=item $self->range
+
+The range for the cell (e.g. "[1:0]") or false (i.e. undef or "") if not
+ranged.
+
 =item $self->submod
 
 Reference to the Verilog::Netlist::Module the cell instantiates.  Only
@@ -269,13 +299,13 @@ Prints debugging information for this cell.
 
 =head1 DISTRIBUTION
 
-Verilog-Perl is part of the L<http://www.veripool.com/> free Verilog EDA
+Verilog-Perl is part of the L<http://www.veripool.org/> free Verilog EDA
 software tool suite.  The latest version is available from CPAN and from
-L<http://www.veripool.com/verilog-perl.html>.
+L<http://www.veripool.org/verilog-perl>.
 
-Copyright 2000-2007 by Wilson Snyder.  This package is free software; you
+Copyright 2000-2019 by Wilson Snyder.  This package is free software; you
 can redistribute it and/or modify it under the terms of either the GNU
-Lesser General Public License or the Perl Artistic License.
+Lesser General Public License Version 3 or the Perl Artistic License Version 2.0.
 
 =head1 AUTHORS
 
