@@ -1,30 +1,19 @@
 # Verilog - Verilog Perl Interface
-# $Id: Port.pm 39061 2007-05-21 14:49:55Z wsnyder $
-# Author: Wilson Snyder <wsnyder@wsnyder.org>
-######################################################################
-#
-# Copyright 2000-2007 by Wilson Snyder.  This program is free software;
-# you can redistribute it and/or modify it under the terms of either the GNU
-# General Public License or the Perl Artistic License.
-# 
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-# 
+# See copyright, etc in below POD section.
 ######################################################################
 
 package Verilog::Netlist::Port;
-use Class::Struct;
 
 use Verilog::Netlist;
 use Verilog::Netlist::Subclass;
+use vars qw($VERSION @ISA);
+use strict;
 @ISA = qw(Verilog::Netlist::Port::Struct
 	Verilog::Netlist::Subclass);
-$VERSION = '2.380';
-use strict;
 
-structs('new',
+$VERSION = '2.468';
+
+structs('_new_base',
 	'Verilog::Netlist::Port::Struct'
 	=>[name     	=> '$', #'	# Name of the port
 	   filename 	=> '$', #'	# Filename this came from
@@ -33,7 +22,7 @@ structs('new',
 	   attributes	=> '%', #'	# Misc attributes for systemperl or other processors
 	   #
 	   direction	=> '$', #'	# Direction (in/out/inout)
-	   type	 	=> '$', #'	# C++ Type (bool/int)
+	   data_type	=> '$', #'	# SystemVerilog Type (logic/integer etc)
 	   comment	=> '$', #'	# Comment provided by user
 	   array	=> '$', #'	# Vectorization
 	   module	=> '$', #'	# Module entity belongs to
@@ -43,23 +32,59 @@ structs('new',
 	   sp_autocreated	=> '$', #'	# Created by /*AUTOINOUT*/
 	   ]);
 
+sub new {
+    my $class = shift;
+    my %params = (@_);
+    $params{data_type} = $params{type} if defined $params{type};  # Backward compatibility
+    if ($params{direction}) {  # Correct common mistakes; plus the parser itself needs this conversion
+	$params{direction} = 'in' if $params{direction} eq 'input';
+	$params{direction} = 'out' if $params{direction} eq 'output';
+    }
+    return $class->_new_base(%params);
+}
+
+sub delete {
+    my $self = shift;
+    my $h = $self->module->_ports;
+    delete $h->{$self->name};
+    return undef;
+}
+
+######################################################################
+
 sub netlist { return $_[0]->module->netlist; }
 
+sub logger { return $_[0]->netlist->logger; }
+
+sub type {  # Backward compatibility only
+    my $self=shift;
+    if ($#_ >= 0) { $self->data_type(@_); }
+    return ($self->data_type || ($self->net && $self->net->type))||'';
+}
+
 sub _link {
     my $self = shift;
     if (!$self->net) {
-	my $net = $self->module->find_net ($self->name);
+	my $net = $self->module->find_net($self->name);
 	if (!$net) {
+	    my $msb;
+	    my $lsb;
+	    if (defined $self->data_type) {
+		$self->data_type =~ /\[([^:]+)(:(.*))?\]$/;
+		$msb = $1;
+		$lsb = defined($3) ? $3 : $1;
+	    }
 	    $net = $self->module->new_net
 		(name=>$self->name,
 		 filename=>$self->filename, lineno=>$self->lineno,
-		 type=>$self->type, array=>$self->array,
-		 comment=>undef,
+		 decl_type=>"port", net_type=>"wire",
+		 data_type=>$self->data_type, array=>$self->array,
+		 comment=>undef, msb=>$msb, lsb=>$lsb,
 		 );
 	    $net->attributes($self->attributes);  # Copy attributes across
 	}
 	if ($net && $net->port && $net->port != $self) {
-	    $self->error ("Port redeclares existing port: ",$self->name,"\n");
+	    $self->error("Port redeclares existing port: ",$self->name,"\n");
 	}
 	$self->net($net);
 	$self->net->port($self);
@@ -80,7 +105,7 @@ sub dump {
     my $self = shift;
     my $indent = shift||0;
     print " "x$indent,"Port:",$self->name(),"  Dir:",$self->direction()
-	,"  Type:",$self->type(),"  Array:",$self->array()||"","\n";
+	,"  DataT:",$self->data_type(),"  Array:",$self->array()||"","\n";
 }
 
 ######################################################################
@@ -99,7 +124,7 @@ Verilog::Netlist::Port - Port for a Verilog Module
   use Verilog::Netlist;
 
   ...
-  my $port = $module->find_port ('pinname');
+  my $port = $module->find_port('pinname');
   print $port->name;
 
 =head1 DESCRIPTION
@@ -115,13 +140,21 @@ See also Verilog::Netlist::Subclass for additional accessors and methods.
 
 =item $self->array
 
-Any array declaration for the port.
+Any array declaration for the port.  This only applies to Verilog 1995
+style ports which can declare port bits independently from the signal
+declarations.  When using Verilog 2001 style ports, see the matching net
+declaration's data_type, msb and lsb methods instead, for example
+C<$module->find_net($port->name)->data_type>.
 
 =item $self->comment
 
 Returns any comments following the definition.  keep_comments=>1 must be
 passed to Verilog::Netlist::new for comments to be retained.
 
+=item $self->data_type
+
+The SystemVerilog data type of the port.
+
 =item $self->direction
 
 The direction of the port: "in", "out", or "inout".
@@ -141,7 +174,8 @@ after the netlist is linked.
 
 =item $self->type
 
-The C++ type of the port.
+Approximately an alias of data_type for backward compatibility.  Do not use
+for new applications.
 
 =back
 
@@ -159,13 +193,13 @@ Prints debugging information for this port.
 
 =head1 DISTRIBUTION
 
-Verilog-Perl is part of the L<http://www.veripool.com/> free Verilog EDA
+Verilog-Perl is part of the L<http://www.veripool.org/> free Verilog EDA
 software tool suite.  The latest version is available from CPAN and from
-L<http://www.veripool.com/verilog-perl.html>.
+L<http://www.veripool.org/verilog-perl>.
 
-Copyright 2000-2007 by Wilson Snyder.  This package is free software; you
+Copyright 2000-2019 by Wilson Snyder.  This package is free software; you
 can redistribute it and/or modify it under the terms of either the GNU
-Lesser General Public License or the Perl Artistic License.
+Lesser General Public License Version 3 or the Perl Artistic License Version 2.0.
 
 =head1 AUTHORS
 
